@@ -1,6 +1,6 @@
 package Plack::Handler::Stomp;
 {
-  $Plack::Handler::Stomp::VERSION = '1.0';
+  $Plack::Handler::Stomp::VERSION = '1.06_01';
 }
 {
   $Plack::Handler::Stomp::DIST = 'Plack-Handler-Stomp';
@@ -16,8 +16,9 @@ use namespace::autoclean;
 use Try::Tiny;
 use Plack::Util;
 
-with 'Net::Stomp::MooseHelpers::CanConnect';
-with 'Net::Stomp::MooseHelpers::CanSubscribe';
+with 'Net::Stomp::MooseHelpers::CanConnect' => { -version => '1.1' };
+with 'Net::Stomp::MooseHelpers::CanSubscribe' => { -version => '1.1' };
+with 'Net::Stomp::MooseHelpers::ReconnectOnFailure';
 
 # ABSTRACT: adapt STOMP to (almost) HTTP, via Plack
 
@@ -50,11 +51,10 @@ has one_shot => (
 sub run {
     my ($self, $app) = @_;
 
-    SERVER_LOOP:
-    while (1) {
-        my $exception;
+    my $exception;
+    $self->reconnect_on_failure(
+        sub {
         try {
-            $self->connect();
             $self->subscribe();
 
             $self->frame_loop($app);
@@ -63,23 +63,24 @@ sub run {
         };
         if ($exception) {
             if (!blessed $exception) {
-                die "unhandled exception $exception";
+                $exception = "unhandled exception $exception";
+                return;
             }
             if ($exception->isa('Plack::Handler::Stomp::Exceptions::AppError')) {
-                die $exception;
-            }
-            if ($exception->isa('Net::Stomp::MooseHelpers::Exceptions::Stomp')) {
-                $self->clear_connection;
-                next SERVER_LOOP;
-            }
-            if ($exception->isa('Plack::Handler::Stomp::Exceptions::OneShot')) {
-                last SERVER_LOOP;
+                return;
             }
             if ($exception->isa('Plack::Handler::Stomp::Exceptions::UnknownFrame')) {
-                die $exception;
+                return;
             }
+            if ($exception->isa('Plack::Handler::Stomp::Exceptions::OneShot')) {
+                $exception=undef;
+                return;
+            }
+            die $exception;
         }
-    }
+    });
+    die $exception if defined $exception;
+    return;
 }
 
 
@@ -88,6 +89,11 @@ sub frame_loop {
 
     while (1) {
         my $frame = $self->connection->receive_frame();
+        if(!$frame || !ref($frame)) {
+            Net::Stomp::MooseHelpers::Exceptions::Stomp->throw({
+                stomp_error => 'empty frame received',
+            });
+        }
         $self->handle_stomp_frame($app, $frame);
 
         Plack::Handler::Stomp::Exceptions::OneShot->throw()
@@ -330,6 +336,7 @@ __PACKAGE__->meta->make_immutable;
 1;
 
 __END__
+
 =pod
 
 =encoding utf-8
@@ -340,7 +347,7 @@ Plack::Handler::Stomp - adapt STOMP to (almost) HTTP, via Plack
 
 =head1 VERSION
 
-version 1.0
+version 1.06_01
 
 =head1 SYNOPSIS
 
@@ -566,4 +573,3 @@ This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
 
 =cut
-
